@@ -1,4 +1,5 @@
 import random
+import numpy as np
 from typing import Iterable, Union, List, Set, Dict
 
 from wsknn.weighting import weight_session_items, weight_item_score
@@ -14,91 +15,132 @@ class WSKNN:
     Parameters
     ----------
     number_of_recommendations : int, default=5
-                                The number of recommended items.
+        The number of recommended items.
 
     number_of_neighbors : int, default=10
-                          The number of closest sessions to choose the items from.
+        The number of the closest sessions to choose the items from.
 
     sampling_strategy : str, default='common_items'
-                        How to filter the initial sample of sessions. Available strategies are:
-                        - 'common_items': sample sessions with the same items as the input session,
-                        - 'recent': sample the most actual sessions,
-                        - 'random': get a random sample of sessions.
+        How to filter the initial sample of sessions. Available strategies are:
+
+        - ``'common_items'``: sample sessions with the same items as the input session,
+        - ``'recent'``: sample the most actual sessions,
+        - ``'random'``: get a random sample of sessions,
+        - ``'weighted_events'``: select sessions based on the specific weights assigned to events.
 
     sample_size : int, default=1000
-                  How many sessions from the model are sampled to make a recommendation.
+        How many sessions from the model are sampled to make a recommendation.
 
     weighting_func : str, default='linear'
-                     The similarity measurement between sessions. Available options: 'linear', 'log' and 'quadratic'.
+        The similarity measurement between sessions. Available options: ``'linear'``, ``'log'`` and ``'quadratic'``.
 
     ranking_strategy : str, default='linear'
-                       How we calculate an item rank (based on its position in a session sequence). Available options
-                       are: 'inv', 'linear', 'log', 'quadratic'.
+        How we calculate an item rank (based on its position in a session sequence). Available options
+        are: ``'inv'``, ``'linear'``, ``'log'``, ``'quadratic'``.
+
+    return_events_from_session : bool, default = True
+        Should algorithm return the same items as in session if there are no neighbors?
 
     required_sampling_event : int or str, default = None
-                              Set this paramater to the event name if sessions with it must be included in
-                              the neighbors selection. For example, this event may be a "purchase".
+        Set this parameter to the event name if sessions with this event must be included in the neighbors' selection.
+        For example, an event name may be the ``"purchase"``.
+
+    required_sampling_event_index : int, default = None
+        If ``required_sampling_event`` parameter is filled then you must pass an index of a row with event names.
+
+    sampling_event_weights_index : int, default = None
+        If ``sampling_strategy`` is set to ``weighted_events`` then you must pass an index of a row with event weights.
+
+    recommend_any : bool, default = False
+        If recommender picks less items than required by the ``number_of_recommendations`` then add random items to
+        the results.
 
     Attributes
     ----------
     weighting_functions: List
-                         List with weighting functions: ['linear', 'log', 'quadratic'].
+        The weighting functions: ``['linear', 'log', 'quadratic']``.
 
     ranking_strategies : List
-                         List with ranking strategies: ['linear', 'log', 'quadratic', 'inv'].
+        The ranking strategies: ``['linear', 'log', 'quadratic', 'inv']``.
 
     session_item_map : Dict
-                       sessions = {
-                           session_id: (
-                               [sequence_of_items],
-                               [sequence_of_timestamps],
-                               [sequence_of_event_type]
-                           )
-                       }
-                       Dict populated with fit() method.
+        The map of items that are occuring the session and their timestamps, and (optional) their types and their
+        weights.
+        >>> sessions = {
+        ...     session_id: (
+        ...     [sequence_of_items],
+        ...     [sequence_of_timestamps],
+        ...     [(optional) event names (types)],
+        ...     [(optional) weights]
+        ...     )
+        ... }
 
     item_session_map : Dict
-                       items = {
-                           item_id: (
-                               [sequence_of_sessions],
-                               [sequence_of_the_first_session_timestamps]
-                           )
-                       }
-                       Dict populated with fit() method.
+        The map of items and the sessions where those items are present, and the first timestamp of those sessions.
+        >>> items = {
+        ...     item_id: (
+        ...     [sequence_of_sessions],
+        ...     [sequence_of_the_first_session_timestamps]
+        ...     )
+        ... }
 
     n_of_recommendations : int
-                           Number of items recommended.
+        The number of items recommended.
 
     number_of_closest_neighbors : int
+        See ``number_of_neighbors`` parameter.
 
     sampling_strategy : str
-                        See sampling_strategy parameter.
+        See ``sampling_strategy`` parameter.
 
     possible_neighbors_sample_size : int
-                                     See sample_size parameter.
+        See ``sample_size`` parameter.
 
     weighting_function : str
-                         See weighting_func parameter.
+        See ``weighting_func`` parameter.
 
     ranking_strategy : str
-                       See ranking_strategy parameter.
+        See ``ranking_strategy`` parameter.
+
+    return_events_from_session : bool, default = True
+        See ``return_events_from_session`` parameter.
 
     required_sampling_event : Union[int, str], default = None
-                              See required_sampling_event parameter.
+        See ``required_sampling_event`` parameter.
+
+    required_sampling_event_index : int, default = None
+        See ``required_sampling_event_index`` parameter.
+
+    sampling_event_weights_index : int, default = None
+        See ``sampling_str_event_weights_index`` parameter.
+
+    recommend_any : bool, default = False
+        See ``recommend_any`` parameter.
 
     Methods
     -------
     fit()
+        Sets input session-items and item-sessions maps.
 
-    predict()
+    recommend()
+        The method predicts the ``n`` next recommendations from a given session.
+
+    set_model_params()
+        Methods resets and maps the new model parameters.
 
     Raises
     ------
-    InvalidDimensionsError : wrong number of nested sequences within session-items or item-sessions maps.
+    InvalidDimensionsError
+        Wrong number of nested sequences within session-items or item-sessions maps.
 
-    InvalidTimestampError : wrong type of timestamp - int is required.
+    InvalidTimestampError
+        Wrong type of timestamp - ``int`` type is required.
 
-    TypeError : wrong type of nested structures within session-items or item-sessions maps.
+    TypeError
+        Wrong type of nested structures within session-items or item-sessions maps.
+
+    IndexError
+        Wrong index of event names or wrong index of event weights.
 
     """
 
@@ -109,9 +151,32 @@ class WSKNN:
                  sample_size: int = 1000,
                  weighting_func: str = 'linear',
                  ranking_strategy: str = 'linear',
-                 required_sampling_event: Union[int, str] = None):
+                 return_events_from_session: bool = True,
+                 required_sampling_event: Union[int, str] = None,
+                 required_sampling_event_index: int = None,
+                 sampling_event_weights_index: int = None,
+                 recommend_any: bool = False):
 
-        self.sampling_strategies = ['common_items', 'recent', 'random']
+        # CHECKS
+
+        # Check if all parameters are given: required sampling event
+
+        if required_sampling_event is not None:
+            # User must provide index of event names
+            if required_sampling_event_index is None:
+                msg = 'With required sampling event given you must provide index of a row with names of events!'
+                raise IndexError(msg)
+
+        # Check if all parameters are given: sampling_strategy == 'weighted_events'
+        if sampling_strategy == 'weighted_events':
+            if sampling_event_weights_index is None:
+                msg = 'If you want to sample sessions based on the weights then you must provide index of ' \
+                      'the row with weights'
+                raise IndexError(msg)
+
+        # INITILIAZATION
+
+        self.sampling_strategies = ['common_items', 'recent', 'random', 'weighted_events']
         self.weighting_functions = ['linear', 'log', 'quadratic']
         self.ranking_strategies = ['linear', 'log', 'quadratic', 'inv']
 
@@ -126,30 +191,43 @@ class WSKNN:
         self.sampling_strategy = self._is_sampling_strategy_valid(sampling_strategy)
         self.weighting_function = self._is_weighting_function_valid(weighting_func)
         self.ranking_strategy = self._is_ranking_strategy_valid(ranking_strategy)
+        self.return_events_from_session = return_events_from_session
+        self.sampling_event_weights_index = sampling_event_weights_index
+        self.required_sampling_event_index = required_sampling_event_index
+        self.recommend_any = recommend_any
 
     # Core methods
 
-    def fit(self, sessions: Dict, items: Dict):
-        """Sets input session-items and item-sessions maps.
+    def fit(self,
+            sessions: Dict,
+            items: Dict):
+        """
+        Sets input session-items and item-sessions maps.
 
         Parameters
         ----------
         sessions : Dict
-                   sessions = {
-                       session_id: (
-                           [ sequence_of_items ],
-                           [ sequence_of_timestamps ],
-                           [ [OPTIONAL] sequence_of_event_types ]
-                       )
-                   }
+            The map of items that are occuring the session and their timestamps, and (optional) their types and their
+            weights.
+
+            >>> sessions = {
+            ...     session_id: (
+            ...     [sequence_of_items],
+            ...     [sequence_of_timestamps],
+            ...     [(optional) event names (types)],
+            ...     [(optional) weights]
+            ...     )
+            ... }
 
         items : Dict
-                items = {
-                    item_id: (
-                        [ sequence_of_sessions ],
-                        [ sequence_of_the_first_session_timestamps ]
-                    )
-                }
+            The map of items and the sessions where those items are present, and the first timestamp of those sessions.
+
+            >>> items = {
+            ...     item_id: (
+            ...     [sequence_of_sessions],
+            ...     [sequence_of_the_first_session_timestamps]
+            ...     )
+            ... }
         """
 
         # Check input data
@@ -159,89 +237,174 @@ class WSKNN:
         self.session_item_map = sessions
         self.item_session_map = items
 
-    def predict(self, sessions: Union[Dict, List],
-                number_of_recommendations=None,
-                number_of_closest_neighbors=None,
-                session_sampling_strategy=None,
-                possible_neighbors_sample_size=None,
-                weighting_strategy=None,
-                rank_strategy=None,
-                required_sampling_event=None) -> Dict:
+    def recommend(self,
+                  event_stream: List,
+                  settings: dict = None) -> List:
         """
         The method predicts n next recommendations from a given session.
 
         Parameters
         ----------
-        sessions : Union[Dict, List]
-                   Sequence of items for recommendation. As a dict, it is a user id (key) and products and
-                   their timestamps (values). As a List it is a nested List of user actions and their timestamps,
-                   and additional properties.
+        event_stream : List
+            Sequence of items for recommendation. It must be a nested ``List`` of lists:
+            >>> [
+            ...     [sequence_of_items],
+            ...     [sequence_of_timestamps],
+            ...     [(optional) event names (types)],
+            ...     [(optional) weights]
+            ... ]
 
-        number_of_recommendations : int or None, default=None
-                                    Resets the number of recommendations.
-
-        number_of_closest_neighbors : int or None, default=None
-                                      Resets the number of closest neighbors.
-
-        session_sampling_strategy : str or None, default=None
-                                    How to filter the initial sample of sessions. Available strategies are:
-                                        - 'common_items': sample sessions with the same items as the input session,
-                                        - 'recent': sample the most actual sessions,
-                                        - 'random': get a random sample of sessions.
-                                        
-        possible_neighbors_sample_size : int or None, default=None
-                                         How many sessions from the model are sampled to make a recommendation. If not
-                                         set then it is the number given during the class initilization.
-
-        weighting_strategy : str or None, default=None
-                             The similarity measurement between sessions. Available options: 'linear', 'log'
-                             and 'quadratic'. If not set then weighting strategy is taken from the parameter set
-                             during the class initialization.
-
-        rank_strategy : str or None, default=None
-                        How we calculate an item rank (based on its position in a session sequence). Available options
-                        are: 'inv', 'linear', 'log', 'quadratic'. If not set then model uses rank strategy given
-                        during the initilization.
-
-        required_sampling_event : int or str, default = None
-                                  Set this paramater to the event name if sessions with it must be included in
-                                  the neighbors selection. For example, this event may be a "purchase".
+        settings : Dict, default = None
+            Model settings and parameters.
 
         Returns
         -------
-        : Dict
-          ranked items in descending order
-          - {user ID: [[item, rank] ...]} (input as a Dict)
-          - {int: [[item, rank] ...]} (input as a List)
+        recommendations : List
+            >>> [
+            ...     (item a, rank a), (item b, rank b)
+            ... ]
         """
 
-        # Set class consts
-        self._reset_model_params(number_of_recommendations,
-                                 number_of_closest_neighbors,
-                                 session_sampling_strategy,
-                                 possible_neighbors_sample_size,
-                                 weighting_strategy,
-                                 rank_strategy,
-                                 required_sampling_event)
+        if settings is not None:
+            self.set_model_params(**settings)
 
-        output_ranks = dict()
-        if isinstance(sessions, dict):
-            for _key, session in sessions.items():
-                recommendations = self._part_predict(session)
-                output_ranks[_key] = recommendations
-        elif isinstance(sessions, list):
-            for idx, session in enumerate(sessions):
-                recommendations = self._part_predict(session)
-                output_ranks[idx] = recommendations
-        else:
-            raise TypeError('Not supported input type for prediction')
-        return output_ranks
+        recommendations = self._predict(event_stream)
 
-    def _part_predict(self, session):
+        return recommendations
+
+    def set_model_params(self,
+                         number_of_recommendations=None,
+                         number_of_neighbors=None,
+                         sampling_strategy=None,
+                         sample_size=None,
+                         weighting_func=None,
+                         ranking_strategy=None,
+                         return_events_from_session=None,
+                         required_sampling_event=None,
+                         recommend_any=False):
+        """
+        Methods resets and maps the new model parameters.
+
+        Parameters
+        ----------
+        number_of_recommendations : int, default = None
+
+        number_of_neighbors : int, default = None
+
+        sampling_strategy : str, default = None
+
+        sample_size : int, default = None
+
+        weighting_func : str, default = None
+
+        ranking_strategy : str, default = None
+
+        return_events_from_session : bool, default = None
+
+        required_sampling_event : str or int, default = None
+
+        recommend_any : bool, default = None
+
+        Raises
+        ------
+        TypeError
+            Wrong input parameter type.
+
+        KeyError
+            Wrong name of sampling strategy, ranking strategy or weighting function.
+
+        """
+        if number_of_recommendations is not None:
+            if isinstance(number_of_recommendations, int):
+                self._set_n_of_recs(number_of_recommendations)
+            else:
+                raise TypeError(f'Number of output recommendations should be an integer, '
+                                f'got {type(number_of_recommendations)} instead')
+
+        if number_of_neighbors is not None:
+            if isinstance(number_of_neighbors, int):
+                self._set_number_of_closest_neighbors(number_of_neighbors)
+            else:
+                raise TypeError(f'Number of closest neighbors should be an integer, '
+                                f'got {type(number_of_neighbors)} instead')
+
+        if sampling_strategy is not None:
+            if isinstance(sampling_strategy, str):
+                session_sampling_strategy = self._is_sampling_strategy_valid(sampling_strategy)
+                self._set_sampling_strategy(session_sampling_strategy)
+            else:
+                raise TypeError(f'Defined sampling strategy should be a string. '
+                                f'Got {type(sampling_strategy)} instead')
+
+        if sample_size is not None:
+            if isinstance(sample_size, int):
+                self._set_possible_neighbors_sample_size(sample_size)
+            else:
+                raise TypeError(f'Number of possible neighbors should be an integer, '
+                                f'got {type(sample_size)} instead')
+
+        if weighting_func is not None:
+            if isinstance(weighting_func, str):
+                weighting_strategy = self._is_weighting_function_valid(weighting_func)
+                self._set_weighting_strategy(weighting_strategy)
+            else:
+                raise TypeError(f'Defined weighting function should be a string. '
+                                f'Got {type(weighting_func)} instead')
+
+        if ranking_strategy is not None:
+            if isinstance(ranking_strategy, str):
+                rank_strategy = self._is_ranking_strategy_valid(ranking_strategy)
+                self._set_ranking_strategy(rank_strategy)
+            else:
+                raise TypeError(f'Defined ranking function should be a string. '
+                                f'Got {type(ranking_strategy)} instead')
+
+        if return_events_from_session is not None:
+            if isinstance(return_events_from_session, bool):
+                self.return_events_from_session = return_events_from_session
+            else:
+                raise TypeError(f'return_events_from_session parameter should be set only to True or False (bool). '
+                                f'Got type {type(return_events_from_session)} instead')
+
+        if required_sampling_event is not None:
+            if isinstance(required_sampling_event, str) or isinstance(required_sampling_event, int):
+                self.required_sampling_event = required_sampling_event
+            else:
+                raise TypeError('Defined required sampling event can be int or str, other datatypes are not supported')
+
+        if recommend_any is not None:
+            if isinstance(recommend_any, bool):
+                self.recommend_any = recommend_any
+            else:
+                raise TypeError(f'recommend_any parameter should be set only to True or False (bool). '
+                                f'Got type {type(recommend_any)} instead')
+
+    def _predict(self, session):
         neighbors = self._nearest_neighbors(session)
-        ranked_items = self._rank_items(neighbors, session)
-        ranked_items.sort(key=lambda x: x[1], reverse=True)
-        recommendations = ranked_items[:self.n_of_recommendations]
+
+        if len(neighbors) == 0:
+            if self.recommend_any:
+                recs = list()
+                recommendations = self._get_more_items(recs)
+                return recommendations
+        else:
+            ranked_items = self._rank_items(neighbors, session)
+            ranked_items.sort(key=lambda x: x[1], reverse=True)
+            recommendations = ranked_items[:self.n_of_recommendations]
+
+            if self.recommend_any:
+                if len(recommendations) < self.n_of_recommendations:
+                    recommendations = self._get_more_items(recommendations)
+
+            return recommendations
+
+    def _get_more_items(self, recommendations):
+        add_items_size = self.n_of_recommendations - len(recommendations)
+        possible_items = list(self.item_session_map.keys())
+        for _ in range(add_items_size):
+            rnd_item = random.choice(possible_items)
+            rnd_rec = (rnd_item, 0.0)
+            recommendations.append(rnd_rec)
         return recommendations
 
     # Settings
@@ -252,13 +415,17 @@ class WSKNN:
         Parameters
         ----------
         sessions : Dict
-                   sessions = {
-                       session_id: (
-                           [ sequence_of_items ],
-                           [ sequence_of_timestamps ],
-                           [ [OPTIONAL] sequence_of_event_type ]
-                       )
-                   }
+            The map of items that are occuring the session and their timestamps, and (optional) their types and their
+            weights.
+
+            >>> sessions = {
+            ...     session_id: (
+            ...     [sequence_of_items],
+            ...     [sequence_of_timestamps],
+            ...     [(optional) event names (types)],
+            ...     [(optional) weights]
+            ...     )
+            ... }
 
         Raises
         ------
@@ -278,10 +445,10 @@ class WSKNN:
         sample_rec = sessions[sample_key]
 
         # Check dimensions
-        test_dims = check_data_dimension(sample_rec, 2) or check_data_dimension(sample_rec, 3)
+        test_dims = check_data_dimension(sample_rec, 2)
 
         if not test_dims:
-            raise InvalidDimensionsError('Session-items map', [2, 3])
+            raise InvalidDimensionsError('Session-items map', 2)
 
         # Check type
         for subrec in sample_rec:
@@ -300,13 +467,15 @@ class WSKNN:
 
         Parameters
         ----------
-        items : dict
-                items = {
-                    item_id: (
-                        [ sequence_of_sessions ],
-                        [ sequence_of_the_first_session_timestamps ]
-                    )
-                }
+        items : Dict
+            The map of items and the sessions where those items are present, and the first timestamp of those sessions.
+
+            >>> items = {
+            ...     item_id: (
+            ...     [sequence_of_sessions],
+            ...     [sequence_of_the_first_session_timestamps]
+            ...     )
+            ... }
 
         Raises
         ------
@@ -353,7 +522,7 @@ class WSKNN:
         Raises
         ------
         KeyError
-            Strategy not in list defined in self.sampling_strategies
+            Strategy not in list defined in ``self.sampling_strategies``
 
         Returns
         -------
@@ -376,7 +545,7 @@ class WSKNN:
         Raises
         ------
         KeyError
-            Strategy not in list ['linear', 'log', 'quadratic', 'inv']
+            Strategy not in list ``['linear', 'log', 'quadratic', 'inv']``
 
         Returns
         strategy : str
@@ -398,7 +567,7 @@ class WSKNN:
         Raises
         ------
         KeyError
-            Weighting function not in list ['linear', 'log', 'quadratic']
+            Weighting function not in list ``['linear', 'log', 'quadratic']``
 
         Returns
         wfunc : str
@@ -410,92 +579,6 @@ class WSKNN:
                   f"Use 'linear', 'log', 'quadratic', instead"
             raise KeyError(msg)
 
-    def _reset_model_params(self,
-                            number_of_recommendations,
-                            number_of_closest_neighbors,
-                            session_sampling_strategy,
-                            possible_neighbors_sample_size,
-                            weighting_strategy,
-                            rank_strategy,
-                            required_sampling_event):
-        """Methods resets and maps new model parameters.
-
-        Parameters
-        ----------
-        number_of_recommendations : int
-
-        number_of_closest_neighbors : int
-
-        session_sampling_strategy : str
-
-        possible_neighbors_sample_size : int
-
-        weighting_strategy : str
-
-        rank_strategy : str
-
-        required_sampling_event : str or int
-
-        Raises
-        -------
-        TypeError
-            Wrong type of an input parameter.
-
-        KeyError
-            Wrong name of sampling strategy, ranking strategy or weighting function.
-
-        """
-        if number_of_recommendations is not None:
-            if isinstance(number_of_recommendations, int):
-                self._set_n_of_recs(number_of_recommendations)
-            else:
-                raise TypeError(f'Number of output recommendations should be an integer, '
-                                f'got {type(number_of_recommendations)} instead')
-
-        if number_of_closest_neighbors is not None:
-            if isinstance(number_of_closest_neighbors, int):
-                self._set_number_of_closest_neighbors(number_of_closest_neighbors)
-            else:
-                raise TypeError(f'Number of closest neighbors should be an integer, '
-                                f'got {type(number_of_closest_neighbors)} instead')
-
-        if session_sampling_strategy is not None:
-            if isinstance(session_sampling_strategy, str):
-                session_sampling_strategy = self._is_sampling_strategy_valid(session_sampling_strategy)
-                self._set_sampling_strategy(session_sampling_strategy)
-            else:
-                raise TypeError(f'Defined sampling strategy should be a string. '
-                                f'Got {type(session_sampling_strategy)} instead')
-
-        if possible_neighbors_sample_size is not None:
-            if isinstance(possible_neighbors_sample_size, int):
-                self._set_possible_neighbors_sample_size(possible_neighbors_sample_size)
-            else:
-                raise TypeError(f'Number of possible neighbors should be an integer, '
-                                f'got {type(possible_neighbors_sample_size)} instead')
-
-        if weighting_strategy is not None:
-            if isinstance(weighting_strategy, str):
-                weighting_strategy = self._is_weighting_function_valid(weighting_strategy)
-                self._set_weighting_strategy(weighting_strategy)
-            else:
-                raise TypeError(f'Defined weighting function should be a string. '
-                                f'Got {type(weighting_strategy)} instead')
-
-        if rank_strategy is not None:
-            if isinstance(rank_strategy, str):
-                rank_strategy = self._is_ranking_strategy_valid(rank_strategy)
-                self._set_ranking_strategy(rank_strategy)
-            else:
-                raise TypeError(f'Defined ranking function should be a string. '
-                                f'Got {type(rank_strategy)} instead')
-
-        if required_sampling_event is not None:
-            if isinstance(required_sampling_event, str) or isinstance(required_sampling_event, int):
-                self.required_sampling_event = required_sampling_event
-            else:
-                raise TypeError('Defined required sampling event can be int or str, other datatypes are not supported')
-
     def _set_n_of_recs(self, n):
         self.n_of_recommendations = n
 
@@ -523,19 +606,20 @@ class WSKNN:
         Parameters
         ----------
         session : List
-                  Session is a sequence of products, events (product view / click) and timestamps of those events:
+            Session is a sequence of products, events (product view / click) and timestamps of those events:
 
-            session = [
-                [sequence_of_items],
-                [sequence_of_timestamps],
-                [sequence_of_event_type]
-            ]
+            >>> session = [
+            ...     [sequence_of_items],
+            ...     [sequence_of_timestamps],
+            ...     [optional sequence_of_event_type],
+            ...     [optional sequence of weights]
+            ... ]
 
         Returns
         -------
-        : List
-          n closest sessions, where n - number of closest sessions or length of ranked session if smaller than
-          number of closest sessions.
+        rank : List
+            The ``n`` closest sessions, where ``n`` - number of the closest sessions or length of ranked session if
+            smaller than number of the closest sessions.
         """
         possible_neighbor_sessions = self._possible_neighbors(session)
         items_sequence = session[0]
@@ -551,18 +635,19 @@ class WSKNN:
         Parameters
         ----------
         session : List
-                  Session is a sequence of products, events (product view / click) and timestamps of those events:
+            Session is a sequence of products, events (product view / click) and timestamps of those events:
 
-            session = [
-                [sequence_of_items],
-                [sequence_of_timestamps],
-                [sequence_of_event_type]
-            ]
+            >>> session = [
+            ...     [sequence_of_items],
+            ...     [sequence_of_timestamps],
+            ...     [optional sequence_of_event_type],
+            ...     [optional sequence of weights]
+            ... ]
 
         Returns
         -------
-        : List
-          Sample of possible neighbors. Sampling controlled by the sampling_strategy attribute.
+        sample_subset : List
+          Sample of possible neighbors. Sampling controlled by the ``sampling_strategy`` attribute.
         """
         session_items = set(session[0])
         common_sessions = set()
@@ -584,15 +669,15 @@ class WSKNN:
         Parameters
         ----------
         closest_neighbors : List
-                            The closest sessions ranked by similarity to a given session.
+            The closest sessions ranked by similarity to a given session.
 
         session : List
-                  User session.
+            User session.
 
         Returns
         -------
-        : List
-          List of rated items in descending order.
+        rank : List
+            The list of rated items in descending order.
         """
         session_items = session[0]
         scores = dict()
@@ -608,9 +693,7 @@ class WSKNN:
                 step = step + 1
 
             for n_item in n_items:
-                if n_item in session_items:
-                    # We do not want to score items viewed by the user
-                    # TODO: condition to control this behavior
+                if n_item in session_items and not self.return_events_from_session:
                     pass
                 else:
                     old_score = scores.get(n_item)
@@ -636,15 +719,15 @@ class WSKNN:
         Parameters
         ----------
         session_items : List
-                        List of items from the customer session.
+            List of items from the customer session.
 
         possible_neighbours : List
-                              List of sessions from the possible neighbors pool (based on a sampling strategy).
+            List of sessions from the possible neighbors pool (based on a sampling strategy).
 
         Returns
         -------
         neighbors : List
-                    List of similar sessions to the customer session.
+            List of similar sessions to the customer session.
         """
 
         pos_weights = dict()
@@ -664,7 +747,7 @@ class WSKNN:
         return neighbours
 
     def _get_sessions_with_event(self, raw_sessions: Set) -> Set:
-        """Method parses available input sessions based on the occurence of a specific event and returns list of
+        """Method parses available input sessions based on the occurrence of a specific event and returns list of
         unique sessions with this event.
 
         Parameters
@@ -674,13 +757,12 @@ class WSKNN:
         Returns
         -------
         unique_session_with_event : Set
-
         """
         new_sessions = set()
 
         for sess in raw_sessions:
             session_sample = self.session_item_map[sess]
-            if self.required_sampling_event in session_sample[2]:
+            if self.required_sampling_event in session_sample[self.required_sampling_event_index]:
                 new_sessions.add(sess)
 
         return new_sessions
@@ -690,16 +772,16 @@ class WSKNN:
 
         Parameters
         ----------
-        sessions : set
-                   Unique sessions.
+        sessions : Set
+            Unique sessions.
 
         session : List
-                  The customer session.
+            The customer session.
 
         Returns
         -------
-        : List
-          List of n possible sessions with the same items as a customer session.
+        result : List
+            List of ``n`` possible sessions with the same items as a customer session.
         """
         rank = [(ses, len(set(self.session_item_map[ses][0]) & set(session[0]))) for ses in sessions]
         rank.sort(key=lambda x: x[1])
@@ -713,15 +795,15 @@ class WSKNN:
         Parameters
         ----------
         all_sessions : set
-                       All sessions to sample possible neighbors.
+            All sessions to sample possible neighbors.
 
         session : List
-                  Customer session.
+            Customer session.
 
         Returns
         -------
-        : List
-          subset of possible neighbors
+        sample : List
+            The subset of possible neighbors
         """
 
         if self.sampling_strategy == 'random':
@@ -730,23 +812,25 @@ class WSKNN:
             return self._sampling_recent(all_sessions)
         elif self.sampling_strategy == 'common_items':
             return self._sampling_common(all_sessions, session)
+        elif self.sampling_strategy == 'weighted_events':
+            return self._sampling_weighted_events(all_sessions)
         else:
             err_msg = f'Defined sampling strategy {self.sampling_strategy} not implemented. Available strategies are:' \
                       f' {self.sampling_strategies}.'
             raise TypeError(err_msg)
 
     def _sampling_random(self, sessions: Set) -> List:
-        """Get random sessions from the sessions space. This method is good to estimate model performance or to test it.
+        """Get random sessions from the sessions space. This method is good to estimate model performance.
 
         Parameters
         ----------
         sessions : set
-                   Unique sessions.
+            Unique sessions.
 
         Returns
         -------
-        : List
-          Random sample of self.possible_neighbors_sample_size sessions.
+        sample : List
+            Random sample of ``self.possible_neighbors_sample_size`` sessions.
         """
 
         sample_size = min(self.possible_neighbors_sample_size, len(sessions))
@@ -760,15 +844,41 @@ class WSKNN:
         Parameters
         ----------
         sessions : set
-                   Unique sessions.
+            Unique sessions.
 
         Returns
         -------
-        : List
-          Most recent sessions. Sample of size possible_neighbors_sample_size.
+        result : List
+            The most recent sessions. Sample of size ``self.possible_neighbors_sample_size``.
         """
         rank = [(sid, self.session_item_map[sid][1]) for sid in sessions]
         rank.sort(key=lambda x: x[1], reverse=True)
         result = [x[0] for x in rank]
         sample_size = min(self.possible_neighbors_sample_size, len(sessions))
         return result[:sample_size]
+
+    def _sampling_weighted_events(self, sessions: Set) -> List:
+        """Get sessions with the highest weights.
+
+        Parameters
+        ----------
+        sessions : set
+            Unique sessions.
+
+        Returns
+        -------
+        result : List
+            Sessions with the highest weights. Sample of size ``self.possible_neighbors_sample_size``.
+        """
+
+        rank = [(
+            ses,
+            np.mean(
+                self.session_item_map[ses][self.sampling_event_weights_index]
+            )
+        ) for ses in sessions]
+
+        rank.sort(key=lambda x: x[1], reverse=True)
+        result = [x[0] for x in rank]
+        sample_size = min(self.possible_neighbors_sample_size, len(sessions))
+        return result[:sample_size]
